import { isArray, isNull, isNumber, isString } from 'lodash';
import { from, Subject } from 'rxjs';
import { forkJoin } from 'rxjs';
const [VALID, INVALID, PENDING, DISABLED] = ['VALID', 'INVALID', 'PENDING', 'DISABLED'];
function find(control, path, delimiter) {
    if (path == null)
        return null;
    if (!Array.isArray(path)) {
        path = path.split(delimiter);
    }
    if (Array.isArray(path) && path.length === 0)
        return null;
    // Not using Array.reduce here due to a Chrome 80 bug
    // https://bugs.chromium.org/p/chromium/issues/detail?id=1049982
    let controlToFind = control;
    path.forEach((name) => {
        if (controlToFind instanceof FormGroup) {
            controlToFind = name in controlToFind.controls ? controlToFind.controls[name] : null;
        }
        else {
            controlToFind = null;
        }
    });
    return controlToFind;
}
function mergeErrors(arrayOfErrors) {
    const res = {};
    arrayOfErrors.forEach((errors) => {
        if (errors != null) {
            Object.assign(res, errors);
        }
    });
    return Object.keys(res).length === 0 ? null : res;
}
function composeValidators(validators) {
    const presentValidators = validators.filter((validator) => validator !== null);
    if (presentValidators.length === 0) {
        return null;
    }
    return function (control) {
        return mergeErrors(presentValidators.map((fn) => fn(control)));
    };
}
function composeAsyncValidators(validators) {
    const presentValidators = validators.filter((validator) => validator !== null);
    if (presentValidators.length === 0) {
        return null;
    }
    return function (control) {
        return new Promise((resolve) => {
            const observables = forkJoin(presentValidators.map((fn) => fn(control)));
            observables.subscribe({
                next: (errors) => {
                    resolve(mergeErrors(errors));
                },
            });
        });
    };
}
function coerceToValidator(validator) {
    return isArray(validator) ? composeValidators(validator) : validator;
}
function coerceToAsyncValidator(asyncValidator) {
    return isArray(asyncValidator) ? composeAsyncValidators(asyncValidator) : asyncValidator;
}
function makeValidatorsArray(validators) {
    if (!validators) {
        return [];
    }
    return isArray(validators) ? validators : [validators];
}
function hasValidator(validators, validator) {
    return Array.isArray(validators) ? validators.includes(validator) : validators === validator;
}
function addValidators(validators, currentValidators) {
    const current = makeValidatorsArray(currentValidators);
    const validatorsToAdd = makeValidatorsArray(validators);
    return current.concat(validatorsToAdd.filter((validator) => !hasValidator(current, validator)));
}
function removeValidators(validators, currentValidators) {
    return makeValidatorsArray(currentValidators).filter((validator) => !hasValidator(validators, validator));
}
export class AbstractControl {
    constructor(validators, asyncValidators) {
        Object.defineProperty(this, "_parent", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: null
        });
        Object.defineProperty(this, "_pristine", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: true
        });
        Object.defineProperty(this, "_errors", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: null
        });
        Object.defineProperty(this, "_hasOwnPendingAsyncValidator", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: false
        });
        Object.defineProperty(this, "_asyncValidationSubscription", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_composedValidatorFn", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_composedAsyncValidatorFn", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_rawValidators", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_rawAsyncValidators", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "asyncVerifyComplete$", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Subject()
        });
        this._rawValidators = validators;
        this._rawAsyncValidators = asyncValidators;
        this._composedValidatorFn = coerceToValidator(this._rawValidators);
        this._composedAsyncValidatorFn = coerceToAsyncValidator(this._rawAsyncValidators);
    }
    get validator() {
        return this._composedValidatorFn;
    }
    set validator(validatorFn) {
        this._rawValidators = this._composedValidatorFn = validatorFn;
    }
    get asyncValidator() {
        return this._composedAsyncValidatorFn;
    }
    set asyncValidator(asyncValidatorFn) {
        this._rawAsyncValidators = this._composedAsyncValidatorFn = asyncValidatorFn;
    }
    get parent() {
        return this._parent;
    }
    get root() {
        let control = this;
        while (control.parent) {
            control = control.parent;
        }
        return control;
    }
    get value() {
        return this._value;
    }
    get status() {
        return this._status;
    }
    get errors() {
        return this._errors;
    }
    get valid() {
        return this._status === VALID;
    }
    get invalid() {
        return this._status === INVALID;
    }
    get pending() {
        return this._status === PENDING;
    }
    get disabled() {
        return this._status === DISABLED;
    }
    get enabled() {
        return this._status !== DISABLED;
    }
    get pristine() {
        return this._pristine;
    }
    get dirty() {
        return !this._pristine;
    }
    setValidators(validators) {
        this._rawValidators = validators;
        this._composedValidatorFn = coerceToValidator(validators);
    }
    setAsyncValidators(validators) {
        this._rawAsyncValidators = validators;
        this._composedAsyncValidatorFn = coerceToAsyncValidator(validators);
    }
    addValidators(validators) {
        this.setValidators(addValidators(validators, this._rawValidators));
    }
    addAsyncValidators(validators) {
        this.setAsyncValidators(addValidators(validators, this._rawAsyncValidators));
    }
    removeValidators(validators) {
        this.setValidators(removeValidators(validators, this._rawValidators));
    }
    removeAsyncValidators(validators) {
        this.setAsyncValidators(removeValidators(validators, this._rawAsyncValidators));
    }
    hasValidator(validator) {
        return hasValidator(this._rawValidators, validator);
    }
    hasAsyncValidator(validator) {
        return hasValidator(this._rawAsyncValidators, validator);
    }
    clearValidators() {
        this.validator = null;
    }
    clearAsyncValidators() {
        this.asyncValidator = null;
    }
    markAsDirty(onlySelf = false) {
        this._pristine = false;
        if (this._parent && !onlySelf) {
            this._parent.markAsDirty(onlySelf);
        }
    }
    markAsPristine(onlySelf = false) {
        this._pristine = true;
        this._forEachChild((control) => {
            control.markAsPristine(true);
        });
        if (this._parent && !onlySelf) {
            this._parent._updatePristine(onlySelf);
        }
    }
    markAsPending(onlySelf = false) {
        this._status = PENDING;
        if (this._parent && !onlySelf) {
            this._parent.markAsPending(onlySelf);
        }
    }
    disable(onlySelf = false) {
        // If parent has been marked artificially dirty we don't want to re-calculate the
        // parent's dirtiness based on the children.
        const skipPristineCheck = this._parentMarkedDirty(onlySelf);
        this._status = DISABLED;
        this._forEachChild((control) => {
            control.disable(true);
        });
        this._updateAncestors(onlySelf, skipPristineCheck);
    }
    enable(onlySelf = false) {
        // If parent has been marked artificially dirty we don't want to re-calculate the
        // parent's dirtiness based on the children.
        const skipPristineCheck = this._parentMarkedDirty(onlySelf);
        this._status = VALID;
        this._forEachChild((control) => {
            control.enable(true);
        });
        this.updateValueAndValidity(true);
        this._updateAncestors(onlySelf, skipPristineCheck);
    }
    setParent(parent) {
        this._parent = parent;
    }
    setErrors(errors) {
        this._errors = errors;
        this._updateControlsErrors();
    }
    updateValueAndValidity(onlySelf = false) {
        this._updateValue();
        if (this.enabled) {
            this._cancelExistingSubscription();
            this._errors = this._runValidator();
            this._status = this._calculateStatus();
            if (this.status === VALID || this.status === PENDING) {
                this._runAsyncValidator();
            }
        }
        if (this._parent && !onlySelf) {
            this._parent.updateValueAndValidity(onlySelf);
        }
    }
    get(path) {
        return find(this, path, '.');
    }
    getError(errorCode, path) {
        const control = path ? this.get(path) : this;
        return control && control.errors ? control.errors[errorCode] : null;
    }
    hasError(errorCode, path) {
        return !!this.getError(errorCode, path);
    }
    _isBoxedValue(formState) {
        return (typeof formState === 'object' &&
            formState !== null &&
            Object.keys(formState).length === 2 &&
            'value' in formState &&
            'disabled' in formState);
    }
    _anyControlsHaveStatus(status) {
        return this._anyControls((control) => control.status === status);
    }
    _anyControlsDirty() {
        return this._anyControls((control) => control.dirty);
    }
    _parentMarkedDirty(onlySelf) {
        return !!(!onlySelf && this._parent && this._parent.dirty && !this._parent._anyControlsDirty());
    }
    _updatePristine(onlySelf = false) {
        this._pristine = !this._anyControlsDirty();
        if (this._parent && !onlySelf) {
            this._parent._updatePristine(onlySelf);
        }
    }
    _updateControlsErrors() {
        this._status = this._calculateStatus();
        if (this._parent) {
            this._parent._updateControlsErrors();
        }
    }
    _calculateStatus() {
        if (this._allControlsDisabled())
            return DISABLED;
        if (this.errors)
            return INVALID;
        if (this._hasOwnPendingAsyncValidator || this._anyControlsHaveStatus(PENDING))
            return PENDING;
        if (this._anyControlsHaveStatus(INVALID))
            return INVALID;
        return VALID;
    }
    _updateAncestors(onlySelf, skipPristineCheck) {
        if (this._parent && !onlySelf) {
            this._parent.updateValueAndValidity(onlySelf);
            if (!skipPristineCheck) {
                this._parent._updatePristine();
            }
        }
    }
    _runValidator() {
        return this.validator ? this.validator(this) : null;
    }
    _cancelExistingSubscription() {
        if (this._asyncValidationSubscription) {
            this._asyncValidationSubscription.unsubscribe();
            this._hasOwnPendingAsyncValidator = false;
        }
    }
    _runAsyncValidator() {
        if (this.asyncValidator) {
            this._status = PENDING;
            this._hasOwnPendingAsyncValidator = true;
            this._asyncValidationSubscription = from(this.asyncValidator(this)).subscribe((errors) => {
                this._hasOwnPendingAsyncValidator = false;
                // This will trigger the recalculation of the validation status, which depends on
                // the state of the asynchronous validation (whether it is in progress or not). So, it is
                // necessary that we have updated the `_hasOwnPendingAsyncValidator` boolean flag first.
                this.setErrors(errors);
                this.asyncVerifyComplete$.next(this);
            });
        }
    }
}
export class FormControl extends AbstractControl {
    constructor(formState = null, validators, asyncValidator) {
        super(validators !== null && validators !== void 0 ? validators : null, asyncValidator !== null && asyncValidator !== void 0 ? asyncValidator : null);
        Object.defineProperty(this, "_value", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_status", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: VALID
        });
        this._applyFormState(formState);
        this.updateValueAndValidity(true);
    }
    setValue(value, onlySelf) {
        this._value = value;
        this.updateValueAndValidity(onlySelf);
    }
    patchValue(value, onlySelf) {
        this.setValue(value, onlySelf);
    }
    reset(formState = null, onlySelf) {
        if (isString(this.value)) {
            formState = '';
        }
        else if (isArray(this.value)) {
            formState = [];
        }
        this._applyFormState(formState);
        this.markAsPristine(onlySelf);
        this.setValue(this.value, onlySelf);
    }
    _updateValue() { }
    _forEachChild() { }
    _anyControls() {
        return false;
    }
    _allControlsDisabled() {
        return this.disabled;
    }
    _applyFormState(formState) {
        if (this._isBoxedValue(formState)) {
            this._value = formState.value;
            formState.disabled ? this.disable(true) : this.enable(true);
        }
        else {
            this._value = formState;
        }
    }
}
export class FormGroup extends AbstractControl {
    constructor(controls, validators, asyncValidator) {
        super(validators !== null && validators !== void 0 ? validators : null, asyncValidator !== null && asyncValidator !== void 0 ? asyncValidator : null);
        Object.defineProperty(this, "controls", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: controls
        });
        Object.defineProperty(this, "_value", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_status", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: VALID
        });
        this._setUpControls();
        this.updateValueAndValidity(true);
    }
    addControl(name, control) {
        if (!(name in this.controls)) {
            this.controls[name] = control;
            control.setParent(this);
        }
        this.updateValueAndValidity();
    }
    removeControl(name) {
        delete this.controls[name];
        this.updateValueAndValidity();
    }
    setControl(name, control) {
        delete this.controls[name];
        this.addControl(name, control);
        this.updateValueAndValidity();
    }
    contains(controlName) {
        return controlName in this.controls;
    }
    setValue(value, onlySelf) {
        this._checkAllValuesPresent(value);
        Object.keys(value).forEach((name) => {
            this._throwIfControlMissing(name);
            this.controls[name].setValue(value[name], true);
        });
        this.updateValueAndValidity(onlySelf);
    }
    patchValue(value, onlySelf) {
        Object.keys(value).forEach((name) => {
            if (this.controls[name]) {
                this.controls[name].patchValue(value[name], true);
            }
        });
        this.updateValueAndValidity(onlySelf);
    }
    reset(value = {}, onlySelf) {
        this._forEachChild((control, name) => {
            control.reset(value[name], true);
        });
        this._updatePristine(onlySelf);
        this.updateValueAndValidity(onlySelf);
    }
    _updateValue() {
        this._value = this._reduceValue();
    }
    _forEachChild(cb) {
        Object.keys(this.controls).forEach((key) => {
            // The list of controls can change (for ex. controls might be removed) while the loop
            // is running (as a result of invoking Forms API in `valueChanges` subscription), so we
            // have to null check before invoking the callback.
            const control = this.controls[key];
            control && cb(control, key);
        });
    }
    _anyControls(condition) {
        for (const controlName of Object.keys(this.controls)) {
            const control = this.controls[controlName];
            if (this.contains(controlName) && condition(control)) {
                return true;
            }
        }
        return false;
    }
    _allControlsDisabled() {
        for (const controlName of Object.keys(this.controls)) {
            if (this.controls[controlName].enabled) {
                return false;
            }
        }
        return Object.keys(this.controls).length > 0 || this.disabled;
    }
    _setUpControls() {
        this._forEachChild((control) => {
            control.setParent(this);
        });
    }
    _reduceValue() {
        return this._reduceChildren({}, (acc, control, name) => {
            acc[name] = control.value;
            return acc;
        });
    }
    _reduceChildren(initValue, fn) {
        let res = initValue;
        this._forEachChild((control, name) => {
            res = fn(res, control, name);
        });
        return res;
    }
    _checkAllValuesPresent(value) {
        this._forEachChild((control, name) => {
            if (value[name] === undefined) {
                throw new Error(`Must supply a value for form control with name: '${name}'.`);
            }
        });
    }
    _throwIfControlMissing(name) {
        if (!Object.keys(this.controls).length) {
            throw new Error(`
        There are no form controls registered with this group yet. If you're using ngModel,
        you may want to check next tick (e.g. use setTimeout).
      `);
        }
        if (!this.controls[name]) {
            throw new Error(`Cannot find form control with name: ${name}.`);
        }
    }
}
export class Validators {
    static min(min) {
        return (control) => {
            if (isString(control.value) || isNumber(control.value)) {
                const value = Number(control.value);
                if (Number.isNaN(value)) {
                    return null;
                }
                else {
                    return value < min ? { min: { min, actual: control.value } } : null;
                }
            }
            return null;
        };
    }
    static max(max) {
        return (control) => {
            if (isString(control.value) || isNumber(control.value)) {
                const value = Number(control.value);
                if (Number.isNaN(value)) {
                    return null;
                }
                else {
                    return value > max ? { max: { max, actual: control.value } } : null;
                }
            }
            return null;
        };
    }
    static required(control) {
        const isEmpty = isArray(control.value) || isString(control.value) ? control.value.length === 0 : isNull(control.value) ? true : false;
        return isEmpty ? { required: true } : null;
    }
    static email(control) {
        if (isString(control.value) &&
            !/^(?=.{1,254}$)(?=.{1,64}@)[a-zA-Z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-zA-Z0-9!#$%&'*+/=?^_`{|}~-]+)*@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$/.test(control.value)) {
            return { email: true };
        }
        return null;
    }
    static minLength(minLength) {
        return (control) => {
            if ((isArray(control.value) || isString(control.value)) && control.value.length < minLength) {
                return { minLength: { minLength, actual: control.value.length } };
            }
            return null;
        };
    }
    static maxLength(maxLength) {
        return (control) => {
            if ((isArray(control.value) || isString(control.value)) && control.value.length > maxLength) {
                return { maxLength: { maxLength, actual: control.value.length } };
            }
            return null;
        };
    }
    static pattern(pattern) {
        if (!pattern)
            return () => null;
        let regex;
        let regexStr;
        if (isString(pattern)) {
            regexStr = '';
            if (pattern.charAt(0) !== '^')
                regexStr += '^';
            regexStr += pattern;
            if (pattern.charAt(pattern.length - 1) !== '$')
                regexStr += '$';
            regex = new RegExp(regexStr);
        }
        else {
            regexStr = pattern.toString();
            regex = pattern;
        }
        return (control) => {
            if ((isString(control.value) || isNumber(control.value)) && !regex.test(control.value.toString())) {
                return { pattern: { pattern: regexStr, actual: control.value } };
            }
            return null;
        };
    }
}
//# sourceMappingURL=form-control.js.map